@@ -4,15 +4,13 @@ import sys
 import csv
 import math
 import glob
-import shutil
-
 import numpy as np
+from dotenv import load_dotenv
 import django
 from django.db import models
 from datetime import datetime
 import matplotlib.pyplot as plt
 from matplotlib.patches import Ellipse
-
 import astropy.units as u
 from astropy.table import Table
 from astropy.io.votable import from_table, writeto
@@ -21,88 +19,92 @@ from astropy.io import fits
 from astropy.wcs import WCS
 from astropy.visualization import PercentileInterval
 from astroquery.skyview import SkyView
-from astropy.utils.data import clear_download_cache
+from wallaby_data_access.utils.io import _write_products, _write_zipped_fits_file
 
 
-Run, Instance, Detection, Product, Source  = None, None, None, None, None
+Run, Instance, Detection, Product, Source = None, None, None, None, None
 SourceDetection, Comment, Tag, TagSourceDetection = None, None, None, None
+Observation, ObservationMetadata, Tile, Postprocessing = None, None, None, None
+KinematicModel = None
 
 
-# utils
-def _write_bytesio_to_file(filename, bytesio):
-    """Write the contents of the given BytesIO to a file.
-    Creates the file or overwrites the file if it does
-    not exist yet. 
-    
+def connect(path="/mnt/shared/wallaby/apps/WALLABY_database"):
+    """Establish connection to database through Django models
+
     """
-    with open(filename, "wb") as outfile:
-        # Copy the BytesIO stream to the output file
-        outfile.write(bytesio.getbuffer())
-            
-
-def _write_zipped_fits_file(filename, product, compress=True):
-    """Compress a .fits file as .fits.gz for a data product.
-   
-    """
-    with io.BytesIO() as buf:
-        buf.write(product)
-        buf.seek(0)
-        if not os.path.isfile(filename):
-            _write_bytesio_to_file(filename, buf)
-            if compress:
-                os.system(f'gzip {filename}')
-
-
-def _write_products(products, prefix):
-    _write_zipped_fits_file('%s_cube.fits' % (prefix), products.cube)
-    _write_zipped_fits_file('%s_chan.fits' % (prefix), products.chan)
-    _write_zipped_fits_file('%s_mask.fits' % (prefix), products.mask)
-    _write_zipped_fits_file('%s_mom0.fits' % (prefix), products.mom0)
-    _write_zipped_fits_file('%s_mom1.fits' % (prefix), products.mom1)
-    _write_zipped_fits_file('%s_mom2.fits' % (prefix), products.mom2)
-
-    # Open spectrum
-    with io.BytesIO() as buf:
-        buf.write(b''.join(products.spec))
-        buf.seek(0)
-        spec_file  = '%s_spec.txt' % (prefix)
-        if not os.path.isfile(spec_file):
-            _write_bytesio_to_file(spec_file, buf)
-
-
-# Connect to WALLABY database
-def connect():
     global Run, Instance, Detection, Product, Source
     global SourceDetection, Comment, Tag, TagSourceDetection
-    os.environ["DJANGO_SECRET_KEY"] = "-=(gyah-@e$-ymbz02mhwu6461zv&1&8uojya413ylk!#bwa-l"
-    os.environ["DJANGO_SETTINGS_MODULE"] = "api.settings"
-    os.environ["DJANGO_ALLOW_ASYNC_UNSAFE"] = "True"
-    os.environ["DATABASE_HOST"] = "146.118.67.204"
-    os.environ["DATABASE_NAME"] = "wallabydb"
-    os.environ["DATABASE_USER"] = "wallaby_user"
-    os.environ["DATABASE_PASSWORD"] = "LKaRsQrNtXZ7vN8L*6"
-    sys.path.append("/mnt/shared/wallaby/apps/SoFiAX_services/api/")
+    global Observation, ObservationMetadata, Tile, Postprocessing
+    global KinematicModel
+    load_dotenv('/mnt/shared/wallaby/apps/WALLABY_data_access/src/.env')
+    sys.path.append(path)
+    sys.path.append(path + "/orm")
     django.setup()
-    from tables.models import Run, Instance, Detection, Product, Source
-    from tables.models import SourceDetection, Comment, Tag, TagSourceDetection
+    from source_finding.models import Run, Instance, Detection, Product, Source
+    from source_finding.models import SourceDetection, Comment, Tag, TagSourceDetection
+    from operations.models import Observation, ObservationMetadata, Tile, Postprocessing
+    from kinematic_model.models import KinematicModel
     return
 
 
-# TODO(austin): get multiple catalogues by tags
+def get_slurm_output(source_name):
+    """Get slurmOutput metadata for a given source
+    TODO(austin): move into metadata submodule
+
+    """
+    source = Source.objects.get(name=source_name)
+    sd = SourceDetection.objects.get(source=source)
+    detection = Detection.objects.get(id=sd.detection_id)
+    run = Run.objects.get(id=detection.run_id)
+    postprocessing = Postprocessing.objects.get(run_id=run.id)
+    tile = Tile.objects.get(identifier=postprocessing.name)
+    obs_A = Observation.objects.get(id=tile.footprint_A.id)
+    obs_B = Observation.objects.get(id=tile.footprint_B.id)
+    meta_A = ObservationMetadata.objects.get(observation=obs_A)
+    meta_B = ObservationMetadata.objects.get(observation=obs_B)
+    return {
+        str(obs_A.sbid): meta_A.__dict__['slurm_output'],
+        str(obs_B.sbid): meta_B.__dict__['slurm_output']
+    }
+
+
+# TODO(austin): some better descriptions
+def get_primary_beam_correction_uses_holography(source_name):
+    """Get beam processing information by parsing slurmOutput
+    TODO(austin): move into metadata submodule
+
+    """
+    holography_keywords = [
+        'ASKAP_PB_TT_IMAGE'.lower(),
+        'ASKAP_PB_CUBE_IMAGE'.lower()
+    ]
+
+    slurm_output = get_slurm_output(source_name)
+    beam_info = {}
+    for k, v in slurm_output.items():
+        use_holography = all([hkw in v.keys() for hkw in holography_keywords])
+        beam_info[k] = use_holography
+    return beam_info
 
 
-# Retrieve catalogue by tag
 def get_catalog(tag):
+    """Get catalogue for a given tag name as an astropy table
+
+    """
     tag = str(tag)
     if tag == "":
-        sys.stderr.write("Please specify a tag to extract a source catalogue, e.g.:\ntable = get_catalog(tag=\"NGC 5044 DR1\")\n")
+        sys.stderr.write(
+            "Please specify a tag to extract a source catalogue, e.g.:\ntable = get_catalog(tag=\"NGC 5044 DR1\")\n"
+        )
         return None
-    
+
     table = Table()
-    
+
     # Get field names
     detection_field_names = [field.name for field in Detection._meta.fields if not isinstance(field, models.ForeignKey)]
     detection_field_names.remove("name")
+    detection_field_names.remove("access_url")
+    detection_field_names.remove("access_format")
     detection_field_names.remove("unresolved")
     detection_field_names.remove("v_opt")
     detection_field_names.remove("v_app")
@@ -123,21 +125,21 @@ def get_catalog(tag):
     detection_field_names.remove("freq_peak")
     source_field_names = [field.name for field in Source._meta.fields if not isinstance(field, models.ForeignKey)]
     source_field_names.remove("id")
-    
+
     # Get sources and detections
     sources = [
         Source.objects.get(id=sd.source_id) for sd in [
-            SourceDetection.objects.get(id=tsd.source_detection_id) for tsd in 
-                TagSourceDetection.objects.filter(tag_id=Tag.objects.get(name=tag).id)
+            SourceDetection.objects.get(id=tsd.source_detection_id) for tsd in
+            TagSourceDetection.objects.filter(tag_id=Tag.objects.get(name=tag).id)
         ]
     ]
     detections = [
         Detection.objects.get(id=sd.detection_id) for sd in [
-            SourceDetection.objects.get(id=tsd.source_detection_id) for tsd in 
-                TagSourceDetection.objects.filter(tag_id=Tag.objects.get(name=tag).id)
+            SourceDetection.objects.get(id=tsd.source_detection_id) for tsd in
+            TagSourceDetection.objects.filter(tag_id=Tag.objects.get(name=tag).id)
         ]
     ]
-    
+
     # Add columns to the table
     for field in source_field_names:
         if field == 'name':
@@ -146,7 +148,7 @@ def get_catalog(tag):
             table[field] = np.array([getattr(s, field) for s in sources], dtype=float)
     for field in detection_field_names:
         table[field] = np.array([getattr(d, field) for d in detections], dtype=float)
-    
+
     # Extract and add comments, if any
     column_comments = []
     for i in range(len(table)):
@@ -155,12 +157,14 @@ def get_catalog(tag):
         for comment in comments:
             column_comments[i].append(comment.comment + " (" + comment.author + ")")
     table.add_column(col=column_comments, name="comments")
-    
+
     # Extract and add tags, if any
     column_tags = []
     for i in range(len(table)):
         column_tags.append([])
-        tags = TagSourceDetection.objects.filter(source_detection_id=SourceDetection.objects.get(detection_id=table["id"][i]))
+        tags = TagSourceDetection.objects.filter(
+            source_detection_id=SourceDetection.objects.get(detection_id=table["id"][i])
+        )
         for t in tags:
             column_tags[i].append(Tag.objects.get(id=t.tag_id).name)
     table.add_column(col=column_tags, name="tags")
@@ -187,7 +191,8 @@ def save_catalog(tag, *args, **kwargs):
 
     """
     table = get_catalog(tag)
-    table.remove_columns(['tags'])
+    table['comments'] = [''.join(comment) for comment in table['comments']]
+    table['tags'] = [''.join(t) for t in table['tags']]
     table.write(*args, **kwargs)
 
 
@@ -199,7 +204,7 @@ def save_products_for_source(tag, source_name, *args, **kwargs):
     try:
         idx = list(table['name']).index(source_name)
         row = table[idx]
-    except Exception as e:
+    except Exception:
         sys.stderr.write("Could not find source with provided name in tagged data.")
         return None
     detection = Detection.objects.get(id=row['id'])
@@ -209,10 +214,10 @@ def save_products_for_source(tag, source_name, *args, **kwargs):
     parent = f'{name}_products'
     if not os.path.isdir(parent):
         os.mkdir(parent)
-    
+
     # Write fits files
     _write_products(products, f'{parent}/{name}')
-    
+
     return
 
 
@@ -246,6 +251,38 @@ def print_tags():
     return
 
 
+# Print list of kinematic model tags
+def get_kinematic_model_tags():
+    tr_tags = [k.team_release_kin for k in KinematicModel.objects.all()]
+    return list(set(tr_tags))
+
+
+def _get_kinematic_model_table(objects):
+    table = Table()
+    columns = [f.name for f in KinematicModel._meta.fields]
+    string_columns = ['team_release', 'team_release_kin']
+    array_columns = ["Rad", "Vrot_model", "e_Vrot_model", "e_Vrot_model_inc", "Rad_SD", "SD_model", "SD_FO_model", "e_SD_model", "e_SD_FO_model_inc"]
+    for field in columns:
+        if field == 'name':
+            table[field] = [getattr(k, field).name for k in objects]
+        elif field in string_columns:
+            table[field] = [getattr(k, field) for k in objects]
+        elif field in array_columns:
+            table[field] = [np.array([v for v in getattr(k, field).split(",")]) for k in objects]
+        elif field == 'QFlag_model':
+            table[field] = np.array([getattr(k, field) for k in objects], dtype=int)
+        else:
+            table[field] = np.array([getattr(k, field) for k in objects], dtype=float)
+    return table
+
+
+
+# Generic function for getting kinematic models as astropy.Table
+def get_kinematic_model(*args, **kwargs):
+    res = KinematicModel.objects.filter(**kwargs)
+    return _get_kinematic_model_table(res)
+
+
 # Retrieve FITS image from database
 def get_image(product):
     with io.BytesIO() as buf:
@@ -282,36 +319,52 @@ def overview_plot(id):
     interval = PercentileInterval(95.0)
     plt.rcParams["figure.figsize"] = (16, 12)
     fig = plt.figure()
-    
+
     # Retrieve products from database
     products = Product.objects.get(detection=id)
-    
+
     # Open moment 0 image
     mom0, header = get_image(products.mom0)
     mom1, header = get_image(products.mom1)
     spectrum = get_spectrum(products.spec)
     wcs = WCS(header)
-    
+
     # Extract coordinate information
     nx = header["NAXIS1"]
     ny = header["NAXIS2"]
-    lon, lat = wcs.all_pix2world(nx/2, ny/2, 0)
-    tmp1, tmp3 = wcs.all_pix2world(0, ny/2, 0)
-    tmp2, tmp4 = wcs.all_pix2world(nx, ny/2, 0)
-    width = np.rad2deg(math.acos(math.sin(np.deg2rad(tmp3)) * math.sin(np.deg2rad(tmp4)) + math.cos(np.deg2rad(tmp3)) * math.cos(np.deg2rad(tmp4)) * math.cos(np.deg2rad(tmp1 - tmp2))))
-    tmp1, tmp3 = wcs.all_pix2world(nx/2, 0, 0)
-    tmp2, tmp4 = wcs.all_pix2world(nx/2, ny, 0)
-    height = np.rad2deg(math.acos(math.sin(np.deg2rad(tmp3)) * math.sin(np.deg2rad(tmp4)) + math.cos(np.deg2rad(tmp3)) * math.cos(np.deg2rad(tmp4)) * math.cos(np.deg2rad(tmp1 - tmp2))))
-    
+    lon, lat = wcs.all_pix2world(nx / 2, ny / 2, 0)
+    tmp1, tmp3 = wcs.all_pix2world(0, ny / 2, 0)
+    tmp2, tmp4 = wcs.all_pix2world(nx, ny / 2, 0)
+    width = np.rad2deg(
+        math.acos(
+            math.sin(np.deg2rad(tmp3)) * math.sin(np.deg2rad(tmp4)) +
+            math.cos(np.deg2rad(tmp3)) * math.cos(np.deg2rad(tmp4)) * math.cos(np.deg2rad(tmp1 - tmp2))
+        )
+    )
+    tmp1, tmp3 = wcs.all_pix2world(nx / 2, 0, 0)
+    tmp2, tmp4 = wcs.all_pix2world(nx / 2, ny, 0)
+    height = np.rad2deg(
+        math.acos(
+            math.sin(np.deg2rad(tmp3)) * math.sin(np.deg2rad(tmp4)) +
+            math.cos(np.deg2rad(tmp3)) * math.cos(np.deg2rad(tmp4)) * math.cos(np.deg2rad(tmp1 - tmp2))
+        )
+    )
+
     # Plot DSS image with HI contours
     try:
         hdu_opt = retrieve_dss_image(lon, lat, width, height)
         wcs_opt = WCS(hdu_opt.header)
-        
+
         bmin, bmax = interval.get_limits(hdu_opt.data)
         ax = plt.subplot(2, 2, 2, projection=wcs_opt)
         ax.imshow(hdu_opt.data, origin="lower")
-        ax.contour(mom0, transform=ax.get_transform(wcs), levels=np.logspace(2.0, 5.0, 10), colors="lightgrey", alpha=1.0)
+        ax.contour(
+            mom0,
+            transform=ax.get_transform(wcs),
+            levels=np.logspace(2.0, 5.0, 10),
+            colors="lightgrey",
+            alpha=1.0
+        )
         ax.grid(color="grey", ls="solid")
         ax.set_xlabel("Right ascension (J2000)")
         ax.set_ylabel("Declination (J2000)")
@@ -319,10 +372,10 @@ def overview_plot(id):
         ax.tick_params(axis="y", which="both", top=False, bottom=False)
         ax.set_title("DSS + Moment 0")
         ax.set_aspect(np.abs(wcs_opt.wcs.cdelt[1] / wcs_opt.wcs.cdelt[0]))
-    except:
+    except Exception:
         sys.stderr.write("Failed to retrieve DSS image.\n")
         pass
-    
+
     # Plot moment 0
     ax2 = plt.subplot(2, 2, 1, projection=wcs)
     ax2.imshow(mom0, origin="lower")
@@ -332,7 +385,7 @@ def overview_plot(id):
     ax2.tick_params(axis="x", which="both", left=False, right=False)
     ax2.tick_params(axis="y", which="both", top=False, bottom=False)
     ax2.set_title("Moment 0")
-    
+
     # Add beam size
     ax2.add_patch(Ellipse((5, 5), 5, 5, 0, edgecolor="grey", facecolor="grey"))
 
@@ -346,10 +399,10 @@ def overview_plot(id):
     ax3.tick_params(axis="x", which="both", left=False, right=False)
     ax3.tick_params(axis="y", which="both", top=False, bottom=False)
     ax3.set_title("Moment 1")
-    
+
     # Plot spectrum
     xaxis = spectrum[1] / 1e+6
-    data  = 1000.0 * np.nan_to_num(spectrum[2])
+    data = 1000.0 * np.nan_to_num(spectrum[2])
     xmin = np.nanmin(xaxis)
     xmax = np.nanmax(xaxis)
     ymin = np.nanmin(data)
@@ -364,10 +417,10 @@ def overview_plot(id):
     ax4.grid(True)
     ax4.set_xlim([xmin, xmax])
     ax4.set_ylim([ymin, ymax])
-    
+
     fig.canvas.draw()
     plt.tight_layout()
-    
+
     return plt
 
 
@@ -378,7 +431,7 @@ def save_overview(tag, *args, **kwargs):
     table = get_catalog(tag)
     parent = '%s_overview' % tag.replace(' ', '_')
     if not os.path.isdir(parent):
-        os.mkdir(parent)    
+        os.mkdir(parent)
 
     for row in table:
         name = row['name'].replace(' ', '_')
@@ -394,8 +447,8 @@ def save_overview(tag, *args, **kwargs):
 def save_spectrum(detection, filename):
     products = Product.objects.get(detection=detection)
     spectrum = products.spec
-    
-    # store spectrum as numpy array     
+
+    # store spectrum as numpy array
     array = []
     with io.BytesIO() as buf:
         buf.write(b''.join(products.spec))
@@ -407,11 +460,11 @@ def save_spectrum(detection, filename):
                 chan, freq, flux, pix = line.strip().split()
                 array.append(np.array([int(chan), float(freq), float(flux), int(pix)]))
     array = np.array(array)
-    
+
     # write to fits file
     t = Table(array, names=('Channel', 'Frequency', 'Flux Density', 'Pixels'), dtype=(int, np.float32, np.float32, int))
     t.write(filename, format='fits')
-    
+
     # update spectrum metadata
     spectrum = fits.open(filename, 'update')
     spectrum[0].data = np.array([[[0]]]).astype('int16')
@@ -468,14 +521,14 @@ def casda_deposit(table, deposit_name):
         detection = Detection.objects.get(id=row['id'])
         products = Product.objects.get(detection=detection)
         filename_prefix = f'{name}_{release}'
-        
+
         # write .fits files
-        _write_zipped_fits_file('%s/%s/%s_cube.fits' % (deposit_name, 'cubelets', filename_prefix), products.cube, compress=False)
-        _write_zipped_fits_file('%s/%s/%s_chan.fits' % (deposit_name, 'moment_maps', filename_prefix), products.chan, compress=False)
-        _write_zipped_fits_file('%s/%s/%s_mask.fits' % (deposit_name, 'cubelets', filename_prefix), products.mask, compress=False)
-        _write_zipped_fits_file('%s/%s/%s_mom0.fits' % (deposit_name, 'moment_maps', filename_prefix), products.mom0, compress=False)
-        _write_zipped_fits_file('%s/%s/%s_mom1.fits' % (deposit_name, 'moment_maps', filename_prefix), products.mom1, compress=False)
-        _write_zipped_fits_file('%s/%s/%s_mom2.fits' % (deposit_name, 'moment_maps', filename_prefix), products.mom2, compress=False)
+        _write_zipped_fits_file('%s/%s/%s_cube.fits' % (deposit_name, 'cubelets', filename_prefix), products.cube, compress=False)  # noqa
+        _write_zipped_fits_file('%s/%s/%s_chan.fits' % (deposit_name, 'moment_maps', filename_prefix), products.chan, compress=False)  # noqa
+        _write_zipped_fits_file('%s/%s/%s_mask.fits' % (deposit_name, 'cubelets', filename_prefix), products.mask, compress=False)  # noqa
+        _write_zipped_fits_file('%s/%s/%s_mom0.fits' % (deposit_name, 'moment_maps', filename_prefix), products.mom0, compress=False)  # noqa
+        _write_zipped_fits_file('%s/%s/%s_mom1.fits' % (deposit_name, 'moment_maps', filename_prefix), products.mom1, compress=False)  # noqa
+        _write_zipped_fits_file('%s/%s/%s_mom2.fits' % (deposit_name, 'moment_maps', filename_prefix), products.mom2, compress=False)  # noqa
         save_spectrum(detection, '%s/%s/%s_spec.fits' % (deposit_name, 'spectra', filename_prefix))
 
         # add source name to each header file
@@ -486,7 +539,7 @@ def casda_deposit(table, deposit_name):
             hdr['WALTR'] = release
             hdul.close()
 
-        # TODO: add SDIBs to WALSBID header cards 
+        # TODO: add SDIBs to WALSBID header cards
 
         # TODO: add moment 0 maps (png) to products under separate folder
         mom0, _ = get_image(products.mom0)
@@ -527,18 +580,18 @@ def casda_deposit(table, deposit_name):
         if field.datatype == 'unicodeChar':
             field.datatype = 'char'
 
-    for resource in votable.resources:        
+    for resource in votable.resources:
         # update rows
         for t in resource.tables:
             for i, row in enumerate(fields):
                 t.fields[i] = row
 
             # add params
-            t.params.append(Param(votable, ID='catalogueName', name='Catalogue Name', value=deposit_name, arraysize='59'))
-            t.params.append(Param(votable, ID='indexedFields', name='Indexed Fields', value='name,ra,dec,freq,f_sum,w20,team_release', arraysize='255'))
-            t.params.append(Param(votable, ID='principalFields', name='Principal Fields', value='name,ra,dec,freq,f_sum,w20,team_release', arraysize='255'))
+            t.params.append(Param(votable, ID='catalogueName', name='Catalogue Name', value=deposit_name, arraysize='59'))  # noqa
+            t.params.append(Param(votable, ID='indexedFields', name='Indexed Fields', value='name,ra,dec,freq,f_sum,w20,team_release', arraysize='255'))  # noqa
+            t.params.append(Param(votable, ID='principalFields', name='Principal Fields', value='name,ra,dec,freq,f_sum,w20,team_release', arraysize='255'))  # noqa
 
     catalogue_file = f'{deposit_name}/catalogue/catalogue.xml'
     writeto(votable, catalogue_file)
 
-    return
\ No newline at end of file
+    return
